import sys, os
from .remote_interfaces import RemoterInterface
from .status import Status

# By AWARE on RC: python3.6 and it happy this *local* version of dataclasses, but ver>3.6 should go with system-wide version
try:
    from dataclasses import dataclass
except:
    from dataclasses_local import dataclass
    
from pathlib     import Path
from typing      import Any

MODULE_AVAILABLE = True
try:
    import paramiko
    from paramiko import SSHClient
except ModuleNotFoundError:
    SSHClient = "SSHClient"
    MODULE_AVAILABLE = False
    # print("SSHHandler is not available") # do we need it?

import subprocess
import select
import socket
import termios
import tty


def viewBar(a,b):
    # original version
    res = a/int(b)*100
    sys.stdout.write(f'\rComplete percent: {res:.2f}%')
    sys.stdout.flush()

def tqdmWrapViewBar(*args, **kwargs):
    try:
        from tqdm import tqdm
    except ImportError:
        # tqdm not installed - construct and return dummy/basic versions
        class Foo():
            @classmethod
            def close(*c):
                pass
        return viewBar, Foo
    else:
        pbar = tqdm(*args, **kwargs)  # make a progressbar
        last = [0]  # last known iteration, start at 0
        def viewBar2(a, b):
            pbar.total = int(b)
            pbar.update(int(a - last[0]))  # update pbar with increment
            last[0] = a  # update last known iteration
        return viewBar2, pbar  # return callback, tqdmInstance


@dataclass
class SSHHandler(RemoterInterface):
    __version__  = "0.0.1"
    internal_name: str       = "[SSHHandler]"
    verbose      : bool      = False
    debug        : bool      = False
    ssh          : SSHClient = None
    #
    host         : str  = "192.168.2.52"
    port         : int  = 22
    username     : str  = "kazakov"
    psw          : str  = ""
    ssh_key      : Path = None

    def __post_init__(self):
        if not MODULE_AVAILABLE: return
        if self.verbose: print(f"===> {self.internal_name} [v{self.__version__}] is initialized.")
        self.local_home  = Path.home()


    # def __str__(self):
    #     return f"SSHHandler(host:{self.host}, username={self.username})"

    # def __del__(self):
    #     self.ssh.close()

    # INTERFACE
    def connect(self):
        if not MODULE_AVAILABLE: return
        if self.verbose: print(f"===> {self.internal_name} is connecting to {self.host}...")

        self.ssh = paramiko.SSHClient()
        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.ssh.load_system_host_keys()
        self.ssh.load_host_keys(filename=str(self.ssh_key))
        self.ssh.connect(
            hostname=self.host, 
            username=self.username, 
            password=self.psw,
            port    =self.port,
            # timeout=10,allow_agent=True,look_for_keys=True,
            # key_filename=str(self.ssh_key)
        )

        self.sftp = self.ssh.open_sftp()
        channel = self.ssh.invoke_shell()
        self.stdin = channel.makefile('wb')
        self.stdout = channel.makefile('rb')
        #interactive.interactive_shell(channel)
        # taking remote_home
        _out   = self.execute("pwd", get_output=True, _silent=True)
        self.remote_home = Path(list(_out)[0].strip())

        if self.verbose: print(f"===> {self.internal_name} is ready! [Remote home: {self.remote_home}]")

    def disconnect(self):
        if not MODULE_AVAILABLE: return
        self.ssh.close()

    def is_connected(self) -> bool:
        if not MODULE_AVAILABLE: return
        try:
            test_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            test_socket.connect((self.host, self.port))
        except Exception as ex:
            print(f"There is a problem: {ex}")
            return False
        else:
            test_socket.close()
        return True

    def download(self, src_fname: Path = None, dst_fname: Path = None):
        if not MODULE_AVAILABLE: return
        return self._get_put_file(src_fname=src_fname, dst_fname=dst_fname, direction="get")

    def upload(self, src_fname: Path = None, dst_fname: Path = None):
        if not MODULE_AVAILABLE: return
        return self._get_put_file(src_fname=src_fname, dst_fname=dst_fname, direction="put")

    def exists(self, path: Path) -> bool:
        if not MODULE_AVAILABLE: return
        path = self._path_checker(path, remote=True)
        return True if self.open_file(src_fname=str(path), _silent=True) else False

    # LOCAL methods
    def get_remote_home(self) -> Path:
        if not MODULE_AVAILABLE: return
        return self.remote_home

    '''
    Be aware that Path.home() will always return local home.
    '''
    def _path_checker(self, _path:Path = Path("~/some/path/2/file.dat"), local:bool = False, remote:bool = False) -> Path:
        if not MODULE_AVAILABLE: return
        if local and remote: raise ValueError

        if local:  correct_home: Path = self.local_home
        if remote: correct_home: Path = self.remote_home

        if _path.home() != correct_home:
            #print("Oh no we need to fix home!")
            _path_without_home: Path = Path(str(_path)[len(str(_path.home()))+1:])
            _path = correct_home / _path_without_home

        return _path

    def open_file(self, src_fname:str = None, _silent=False) -> object:
        if not MODULE_AVAILABLE: return
        link2file = None
        if not src_fname: 
            print("Please, provide src_fname: '~/my/path/to/file.dat' ")
        else:
            src_fname = self._path_checker(src_fname, remote=True)
            try:
                link2file = self.sftp.open(str(src_fname), mode='r')
            except FileNotFoundError:
                if not _silent: print(f"{self.internal_name} {str(src_fname)} not exists...")
        return link2file

    def _get_put_file(self, src_fname:Path = None, dst_fname:Path = None, direction:str ="get/put",):
        if not MODULE_AVAILABLE: return
        if src_fname is None or dst_fname is None:
            print("Please, provide  src_fname: '~/path/to/file.dat' and dst_fname: '~/path/to/file2.dat'.")
        else:
            if direction.lower()=="get":
                src_fname = self._path_checker(src_fname, remote=True)
                dst_fname = self._path_checker(dst_fname, local=True)
            if direction.lower()=="put":
                src_fname = self._path_checker(src_fname, local=True)
                dst_fname = self._path_checker(dst_fname, remote=True)

            try:
                cbk, pbar = tqdmWrapViewBar(ascii=True, unit='b', unit_scale=True)
                try:
                    if direction.lower()=="get":
                        self.sftp.get(remotepath = str(src_fname), localpath  = str(dst_fname), callback=cbk)
                    if direction.lower()=="put":
                        self.sftp.put(remotepath = str(dst_fname), localpath  = str(src_fname), callback=cbk)
                    #file_channel = self.sftp.get_channel()
                    #print("STATUS:", file_channel.recv_exit_status())
                    pbar.close()
                    print()
                    return Status("OK") # ["OK", 0, 0,]
                except IOError   as e:  return Status(str(e)) # ["IOERROR: " + str(e),0,0]
                except Exception as e:  return Status(str(e)) #["Error: "   + str(e),0,0]
            except FileNotFoundError:   print(f"{self.internal_name} {self.remote_home/src_fname} not found.")

    """
    Execute command:

    PTY: use bash as enviroment for the call: typically not used.
    """
    def execute(self, cmd:str = "pwd", get_output:bool = False, _silent: bool = False, pty=False) -> Any:
        if not MODULE_AVAILABLE: return
        if pty: cmd = f"bash -l -c '{cmd}'"
        stdin_, stdout_, stderr_ = self.ssh.exec_command(cmd)
        stdout_.channel.recv_exit_status()
        if get_output: output = []

        out_lines = stdout_.readlines()
        err_lines = stderr_.readlines()
        for line in [line.strip() for line in out_lines]:
            if not _silent: print(f"->>: {line}")
            if get_output: output.append(line.strip())
        if err_lines:
            for line in [line.strip() for line in err_lines]:
                if not _silent: print(f"-?>: {line}")
                if get_output: output.append(line.strip())
        #return stdin_, stdout_, stderr_
        if get_output: return output

    """
    WARNING: Experimental use it on your own risk
    """
    def _open_shell(self):
        if not MODULE_AVAILABLE: return
        # get the current TTY attributes to reapply after
        # the remote shell is closed
        oldtty_attrs = termios.tcgetattr(sys.stdin)

        # invoke_shell with default options is vt100 compatible
        # which is exactly what you want for an OpenSSH imitation
        #channel = connection.invoke_shell()
        channel = self.ssh.invoke_shell()

        def resize_pty():
            # resize to match terminal size
            tty_height, tty_width = \
                    subprocess.check_output(['stty', 'size']).split()

            # try to resize, and catch it if we fail due to a closed connection
            try:
                channel.resize_pty(width=int(tty_width), height=int(tty_height))
            except paramiko.ssh_exception.SSHException:
                pass

        # wrap the whole thing in a try/finally construct to ensure
        # that exiting code for TTY handling runs
        try:
            stdin_fileno = sys.stdin.fileno()
            tty.setraw(stdin_fileno)
            tty.setcbreak(stdin_fileno)

            channel.settimeout(0.0)

            is_alive = True

            while is_alive:
                # resize on every iteration of the main loop
                resize_pty()

                # use a unix select call to wait until the remote shell
                # and stdin are ready for reading
                # this is the block until data is ready
                read_ready, write_ready, exception_list = \
                        select.select([channel, sys.stdin], [], [])

                # if the channel is one of the ready objects, print
                # it out 1024 chars at a time
                if channel in read_ready:
                    # try to do a read from the remote end and print to screen
                    try:
                        out = channel.recv(1024)

                        # remote close
                        if len(out) == 0:
                            is_alive = False
                        else:
                            # rely on 'print' to correctly handle encoding
                            print(out, end='')
                            sys.stdout.flush()

                    # do nothing on a timeout, as this is an ordinary condition
                    except socket.timeout:
                        pass

                # if stdin is ready for reading
                if sys.stdin in read_ready and is_alive:
                    # send a single character out at a time
                    # this is typically human input, so sending it one character at
                    # a time is the only correct action we can take

                    # use an os.read to prevent nasty buffering problem with shell
                    # history
                    char = os.read(stdin_fileno, 1)

                    # if this side of the connection closes, shut down gracefully
                    if len(char) == 0:
                        is_alive = False
                    else:
                        channel.send(char)

            # close down the channel for send/recv
            # this is an explicit call most likely redundant with the operations
            # that caused an exit from the REPL, but unusual exit conditions can
            # cause this to be reached uncalled
            channel.shutdown(2)

        # regardless of errors, restore the TTY to working order
        # upon exit and print that connection is closed
        finally:
            termios.tcsetattr(sys.stdin, termios.TCSAFLUSH, oldtty_attrs)
            print(f"Paramiko channel to {self.host} closed.")


